// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "assert/assert.hpp"

#define ASSET_LOADER_BUFFER_SIZE 32768

#if !defined(DAWN_ASSET_LOCATION)
  #error Asset Archive Location has not been defined.
#endif

namespace Dawn {
  class IAssetLoader {
    protected:
      std::string fileName;

    public:
      /**
       * Create a new asset loader. Asset Loaders can be used to load data from
       * a file in a myriad of ways.
       * 
       * @param fileName File name of the asset that is to be loaded.
       */
      IAssetLoader(std::string fileName);

      /**
       * Platform-centric method to open a file buffer to an asset.
       */
      virtual void open() = 0;

      /**
       * Closes the previously ppened asset.
       * @return 0 if successful, otherwise false.
       */
      virtual int32_t close() = 0;

      /**
       * Read bytes from buffer.
       * @param buffer Pointer to a ubyte array to buffer data into.
       * @param size Length of the data buffer (How many bytes to read).
       * @return The count of bytes read.
       */
      virtual size_t read(uint8_t *buffer, size_t size) = 0;

      /**
       * Get the size of the asset.
       * @return The size of the asset in bytes.
       */
      virtual size_t getSize() = 0;

      /**
       * Skips the read head forward to a given position.
       * 
       * @param n Count of bytes to progress the read head by.
       * @return Count of bytes progressed.
       */
      virtual size_t skip(size_t n) = 0;

      /**
       * Rewind the read head to the beginning of the file.
       */
      virtual void rewind() = 0;

      /**
       * Sets the absolute position of the read head within the buffer of the
       * file.
       * 
       * @param absolutePosition Absolute position to set the read head to.
       */
      size_t setPosition(size_t absolutePosition);

      /**
       * Returns the current position of the read head.
       * 
       * @return The current read head position.
       */
      virtual size_t getPosition() = 0;

      /**
       * Run a callback for each byte within the asset. The callback will
       * receive each byte individually.
       * 
       * @tparam T Type of instance to run callback against.
       * @param instance Instance of the object to run the callback against.
       * @param callback Callback method on the class to run the callback for.
       * @return The count of bytes read.
       */
      template<class T>
      size_t loadBufferedCallback(T *instance, bool_t (T::*callback)(uint8_t n)) {
        uint8_t buffer[1024];
        size_t read, length;
        int32_t i;
        bool_t result;

        assertNotNull(instance, "Instance cannot be null.");
        assertNotNull(callback, "Callback cannot be null.");

        // Open the buffer.
        this->open();

        // Reset length size
        length = 0;

        // Buffer from input
        while((read = this->read(buffer, 1024)) != 0) {
          for(i = 0; i < read; i++) {
            result = ((*instance).*(callback))(buffer[i]);
            if(!result) {
              length += i;
              break;
            }
          }
          if(!result) break;
          length += read;
        }

        // Close the buffer
        this->close();

        return length;
      }

      /**
       * Cleanup the asset loader.
       */
      virtual ~IAssetLoader();
  };
}