// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "TrueTypeFaceTexture.hpp"
#include "util/memory.hpp"

using namespace Dawn;

TrueTypeFaceTexture::TrueTypeFaceTexture(
  FT_Face face,
  struct TrueTypeFaceTextureStyle style
) {
  assertTrue(style.fontSize < 256, "TrueTypeFaceTexture::TrueTypeFaceTexture: Font size cannot be greater than 256");

  this->face = face;
  this->style = style;

  // Set freetype font size prior to baking.
  if(FT_Set_Pixel_Sizes(face, 0, style.fontSize)) {
    assertUnreachable("TrueTypeFaceTexture::TrueTypeFaceTexture: Failed to set font size");
  }

  size_t w = 0, h = 0;
  FT_ULong c;

  // First pass, determine the textures' dimensions.
  for(c = TRUE_TYPE_CHAR_BEGIN; c < TRUE_TYPE_CHAR_END; c++) {
    // Load the character
    auto ret = FT_Load_Char(face, c, ~FT_LOAD_RENDER);
    if(ret) {
      assertUnreachable("TrueTypeFaceTexture::TrueTypeFaceTexture: Failed to load character (0)");
    }

    if(face->glyph->bitmap.width == 0 || face->glyph->bitmap.rows == 0) continue;

    // Update the width and height
    w = mathMax<size_t>(w, face->glyph->bitmap.width);
    h += face->glyph->bitmap.rows;
  }

  assertTrue(w > 0, "TrueTypeFaceTexture::TrueTypeFaceTexture: Width cannot be less than or equal to 0");
  assertTrue(h > 0, "TrueTypeFaceTexture::TrueTypeFaceTexture: Height cannot be less than or equal to 0");

  // Now buffer pixels to the texture
  float_t y = 0;

  // I'd love to just buffer straight to the GPU, but it seems that is a bit 
  // unstable right now.
  uint8_t *buffer = (uint8_t *)memoryAllocateEmpty(w * h, sizeof(uint8_t));

  size_t offset = 0;
  struct TrueTypeCharacter info;
  for(c = TRUE_TYPE_CHAR_BEGIN; c < TRUE_TYPE_CHAR_END; c++) {
    // Load the character
    if(FT_Load_Char(face, c, FT_LOAD_RENDER)) {
      assertUnreachable("TrueTypeFaceTexture::TrueTypeFaceTexture: Failed to load character (1)");
    }

    // Store the character information
    info.advanceX = (float_t)(face->glyph->advance.x >> 6);
    info.advanceY = (float_t)(face->glyph->advance.y >> 6);
    info.bitmapSize = glm::vec2(face->glyph->bitmap.width, face->glyph->bitmap.rows);
    info.bitmapPosition = glm::vec2(face->glyph->bitmap_left, -face->glyph->bitmap_top);
    info.textureY = y;
    char c2 = (char)c;
    this->characterData[c] = info;

    // Buffer the pixels, oh dear GOD there has to be a more efficient way.
    for(int32_t i = 0; i < face->glyph->bitmap.rows; i++) {
      memoryCopy(
        (void *)(face->glyph->bitmap.buffer + (i * face->glyph->bitmap.width)),
        (void *)(buffer + offset),
        face->glyph->bitmap.width * sizeof(uint8_t)
      );
      offset += w * sizeof(uint8_t);
      assertTrue(offset <= (w * h * sizeof(uint8_t)), "TrueTypeFaceTexture::TrueTypeFaceTexture: Buffer overflow");
    }
    y += face->glyph->bitmap.rows;
  }

  this->texture.setSize(w, h, TEXTURE_FORMAT_R, TEXTURE_DATA_FORMAT_UNSIGNED_BYTE);
  this->texture.buffer(buffer);
  memoryFree(buffer);
}

struct TrueTypeCharacter TrueTypeFaceTexture::getCharacterData(FT_ULong c) {
  return this->characterData[c];
}

TrueTypeFaceTexture::~TrueTypeFaceTexture() {
  FT_Done_Face(this->face);
}